import argparse
import logging
import logging.config
import os
import pathlib
from logging.handlers import RotatingFileHandler
from typing import Optional

from robotcode.core.logging import LoggingDescriptor
from robotcode.core.utils.debugpy import start_debugpy

from .__version__ import __version__

_logger = LoggingDescriptor(name=__package__)


def get_log_handler(logfile: str) -> logging.FileHandler:
    log_fn = pathlib.Path(logfile)
    roll_over = log_fn.exists()

    handler = RotatingFileHandler(log_fn, backupCount=5)
    formatter = logging.Formatter(
        fmt="[%(levelname)-7s] %(asctime)s (%(name)s) %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    handler.setFormatter(formatter)

    if roll_over:
        handler.doRollover()

    return handler


def run_server(mode: str, port: int, pipe_name: Optional[str]) -> int:
    from robotcode.jsonrpc2.server import JsonRpcServerMode, TcpParams

    from .robotframework.server import RobotLanguageServer

    with RobotLanguageServer(
        mode=JsonRpcServerMode(mode),
        tcp_params=TcpParams("127.0.0.1", port),
        pipe_name=pipe_name,
    ) as server:
        try:
            server.run()
        except SystemExit:
            raise
        except KeyboardInterrupt:
            return 1
        except BaseException as e:
            _logger.exception(e)
            return 1

        return 0


def create_parser() -> argparse.ArgumentParser:
    result = argparse.ArgumentParser(
        description="RobotCode Language Server",
        prog=__package__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    result.add_argument("--version", action="store_true", help="shows the version and exits")

    result.add_argument(
        "-m",
        "--mode",
        default="stdio",
        choices=["stdio", "pipe", "socket", "tcp"],
        help="communication mode",
    )

    result.add_argument(
        "--stdio",
        action="store_true",
        help="run in stdio mode (Shortcut for --mode=stdio)",
    )
    result.add_argument(
        "--socket",
        default=None,
        help="run in socket mode (Shortcut for --mode=socket --port ...)",
        type=int,
    )
    result.add_argument(
        "--pipe",
        default=None,
        help="run in named pipe mode (Shortcut for --mode=pipe --pipe-name ...)",
        type=str,
    )

    result.add_argument(
        "-p",
        "--port",
        default=6610,
        help="server listen port (mode socket and tcp)",
        type=int,
    )

    result.add_argument("-pn", "--pipe-name", default=None, help="pipe name for pipe mode", type=str)

    result.add_argument("--log", action="store_true", help="enable logging")
    result.add_argument("--log-all", action="store_true", help="enable all log messages")
    result.add_argument("--log-json-rpc", action="store_true", help="show json-rpc log messages")
    result.add_argument(
        "--log-json-rpc-data",
        action="store_true",
        help="show json-rpc-data log messages",
    )
    result.add_argument(
        "--log-language-server",
        action="store_true",
        help="show language server log messages",
    )
    result.add_argument(
        "--log-language-server-parts",
        action="store_true",
        help="show language server parts log messages",
    )
    result.add_argument(
        "--log-robotframework",
        action="store_true",
        help="show robotframework language server log messages",
    )
    result.add_argument(
        "--debug-asyncio",
        action="store_true",
        help="enable async io debugging messages",
    )
    result.add_argument("--log-asyncio", action="store_true", help="show asyncio log messages")
    result.add_argument(
        "--log-config",
        default=None,
        help="reads logging configuration from file",
        metavar="FILE",
    )
    result.add_argument("--log-file", default=None, help="enables logging to file", metavar="FILE")
    result.add_argument(
        "--log-level",
        default="WARNING",
        help="sets the overall log level",
        metavar="LEVEL",
    )
    result.add_argument(
        "--call-tracing",
        action="store_true",
        help="enables log tracing of method calls",
    )
    result.add_argument(
        "--call-tracing-default-level",
        default="TRACE",
        help="sets the default level for call tracing",
        metavar="LEVEL",
    )
    result.add_argument("--debugpy", action="store_true", help="starts a debugpy session")
    result.add_argument(
        "--debugpy-port",
        default=5678,
        help="sets the port for debugpy session",
        type=int,
        metavar="PORT",
    )
    result.add_argument(
        "--debugpy-wait-for-client",
        action="store_true",
        help="waits for debugpy client to connect",
    )
    return result


DEFAULT_LOG_LEVEL = logging.ERROR


def init_logging(args: argparse.Namespace) -> None:
    log_level = logging._checkLevel(args.log_level) if args.log else logging.WARNING  # type: ignore

    logging.basicConfig(level=log_level, format="%(name)s:%(levelname)s: %(message)s")

    if args.log_file is not None:
        _logger.logger.addHandler(get_log_handler(args.log_file))

    if not args.log_all:
        if not args.log_robotframework:
            logging.getLogger("robotcode.language_server.robotframework").setLevel(DEFAULT_LOG_LEVEL)
        if not args.log_language_server_parts:
            logging.getLogger("robotcode.language_server.common.parts").setLevel(DEFAULT_LOG_LEVEL)
        if not args.log_language_server:
            logging.getLogger("robotcode.language_server.common").setLevel(DEFAULT_LOG_LEVEL)
        if not args.log_json_rpc:
            logging.getLogger("robotcode.jsonrpc2").setLevel(DEFAULT_LOG_LEVEL)
        if not args.log_json_rpc_data:
            logging.getLogger("robotcode.jsonrpc2.protocol.JsonRPCProtocol_data").setLevel(DEFAULT_LOG_LEVEL)
        if not args.log_asyncio:
            logging.getLogger("asyncio").setLevel(DEFAULT_LOG_LEVEL)


def main() -> int:
    parser = create_parser()

    args = parser.parse_args()

    if args.version:
        print(__version__)
        return 251  # 251 is the exit code for --version

    if args.call_tracing:
        LoggingDescriptor.set_call_tracing(True)
    if args.call_tracing_default_level:
        LoggingDescriptor.set_call_tracing_default_level(
            logging._checkLevel(args.call_tracing_default_level)  # type: ignore
        )

    if args.debug_asyncio:
        os.environ["PYTHONASYNCIODEBUG"] = "1"
        logging.getLogger("asyncio").level = logging.DEBUG
    else:
        logging.getLogger("asyncio").level = DEFAULT_LOG_LEVEL

    if args.log:
        if args.log_config is not None:
            if not os.path.exists(args.log_config):
                raise FileNotFoundError(f"Log-config file '{args.log_config}' not exists.")

            logging.config.fileConfig(args.log_config, disable_existing_loggers=True)
        else:
            init_logging(args)

    _logger.info(lambda: f"starting language server version={__version__}")
    _logger.debug(lambda: f"args={args}")

    if args.debugpy:
        start_debugpy(args.debugpy_port, args.debugpy_wait_for_client)

    mode = args.mode

    if args.stdio:
        mode = "stdio"

    port = args.port

    if args.socket is not None:
        port = args.socket
        mode = "socket"

    pipe_name = args.pipe_name
    if args.pipe is not None:
        mode = "pipe"
        pipe_name = args.pipe

    return run_server(mode, port, pipe_name)



from __future__ import annotations

import sys
from os import PathLike
from pathlib import Path
from typing import Optional, Union


def find_file(
    path: Union[Path, PathLike[str], str],
    basedir: Union[Path, PathLike[str], str] = ".",
    file_type: Optional[str] = None,
) -> str:
    return find_file_ex(path, basedir, file_type)


def find_file_ex(
    path: Union[Path, PathLike[str], str],
    basedir: Union[Path, PathLike[str], str] = ".",
    file_type: Optional[str] = None,
) -> str:
    from robot.errors import DataError

    path = Path(path)
    ret = _find_absolute_path(path) if path.is_absolute() else _find_relative_path(path, basedir)
    if ret:
        return str(ret)
    default = file_type or "File"

    file_type = (
        {"Library": "Test library", "Variables": "Variable file", "Resource": "Resource file"}.get(file_type, default)
        if file_type
        else default
    )

    raise DataError("%s '%s' does not exist." % (file_type, path))


def _find_absolute_path(path: Union[Path, PathLike[str], str]) -> Optional[str]:
    if _is_valid_file(path):
        return str(path)
    return None


def _find_relative_path(
    path: Union[Path, PathLike[str], str],
    basedir: Union[Path, PathLike[str], str],
) -> Optional[str]:
    for base in [basedir, *sys.path]:
        if not base:
            continue
        base_path = Path(base)

        if not base_path.is_dir():
            continue

        ret = Path(base, path).absolute()

        if _is_valid_file(ret):
            return str(ret)
    return None


def _is_valid_file(path: Union[Path, PathLike[str], str]) -> bool:
    path = Path(path)
    return path.is_file() or (path.is_dir() and Path(path, "__init__.py").is_fifo())



import sys
from pathlib import Path

__file__ = str(Path(__file__).absolute())
if __file__.endswith((".pyc", ".pyo")):
    __file__ = __file__[:-1]

if __name__ == "__main__" and __package__ is None or __package__ == "":
    file = Path(__file__).resolve()
    parent, top = file.parent, file.parents[2]

    if str(top) not in sys.path:
        sys.path.append(str(top))

    try:
        sys.path.remove(str(parent))
    except ValueError:
        pass

    __package__ = "robotcode.language_server"

from .cli import main

if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional

from robotcode.core.logging import LoggingDescriptor
from robotcode.language_server.common.decorators import language_id
from robotcode.language_server.common.lsp_types import DocumentHighlight, DocumentHighlightKind, Position
from robotcode.language_server.common.text_document import TextDocument

if TYPE_CHECKING:
    from robotcode.language_server.robotframework.protocol import RobotLanguageServerProtocol

from .model_helper import ModelHelperMixin
from .protocol_part import RobotLanguageServerProtocolPart


class RobotDocumentHighlightProtocolPart(RobotLanguageServerProtocolPart, ModelHelperMixin):
    _logger = LoggingDescriptor()

    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)

        parent.document_highlight.collect.add(self.collect)

    @language_id("robotframework")
    @_logger.call
    async def collect(
        self,
        sender: Any,
        document: TextDocument,
        position: Position,
    ) -> Optional[List[DocumentHighlight]]:
        namespace = await self.parent.documents_cache.get_namespace(document)

        all_variable_refs = await namespace.get_variable_references()
        if all_variable_refs:
            for var, var_refs in all_variable_refs.items():
                for r in var_refs:
                    if (var.source == namespace.source and position in var.name_range) or position in r.range:
                        return [
                            *(
                                [DocumentHighlight(var.name_range, DocumentHighlightKind.TEXT)]
                                if var.source == namespace.source
                                else []
                            ),
                            *(DocumentHighlight(e.range, DocumentHighlightKind.TEXT) for e in var_refs),
                        ]

        all_kw_refs = await namespace.get_keyword_references()
        if all_kw_refs:
            for kw, kw_refs in all_kw_refs.items():
                for r in kw_refs:
                    if (kw.source == namespace.source and position in kw.range) or position in r.range:
                        return [
                            *(
                                [DocumentHighlight(kw.range, DocumentHighlightKind.TEXT)]
                                if kw.source == namespace.source
                                else []
                            ),
                            *(DocumentHighlight(e.range, DocumentHighlightKind.TEXT) for e in kw_refs),
                        ]

        return None



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Optional

from robotcode.core.async_itertools import async_next
from robotcode.core.async_tools import run_coroutine_in_thread, threaded
from robotcode.core.dataclasses import CamelSnakeMixin
from robotcode.core.logging import LoggingDescriptor
from robotcode.jsonrpc2.protocol import rpc_method
from robotcode.language_server.common.lsp_types import Position, Range, TextDocumentIdentifier
from robotcode.language_server.robotframework.utils.ast_utils import (
    HasTokens,
    get_nodes_at_position,
    get_tokens_at_position,
    range_from_token,
)

from .model_helper import ModelHelperMixin
from .protocol_part import RobotLanguageServerProtocolPart

if TYPE_CHECKING:
    from robotcode.language_server.robotframework.protocol import RobotLanguageServerProtocol


@dataclass(repr=False)
class EvaluatableExpressionParams(CamelSnakeMixin):
    text_document: TextDocumentIdentifier
    position: Position


@dataclass(repr=False)
class EvaluatableExpression(CamelSnakeMixin):
    range: Range
    expression: Optional[str]


class RobotDebuggingUtilsProtocolPart(RobotLanguageServerProtocolPart, ModelHelperMixin):
    _logger = LoggingDescriptor()

    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)

    @rpc_method(name="robot/debugging/getEvaluatableExpression", param_type=EvaluatableExpressionParams)
    @threaded()
    @_logger.call
    async def _get_evaluatable_expression(
        self,
        text_document: TextDocumentIdentifier,
        position: Position,
        *args: Any,
        **kwargs: Any,
    ) -> Optional[EvaluatableExpression]:
        from robot.parsing.lexer.tokens import Token as RobotToken

        async def run() -> Optional[EvaluatableExpression]:
            document = await self.parent.documents.get(text_document.uri)
            if document is None:
                return None

            namespace = await self.parent.documents_cache.get_namespace(document)
            model = await self.parent.documents_cache.get_model(document, False)

            nodes = await get_nodes_at_position(model, position)
            node = nodes[-1]

            if not isinstance(node, HasTokens):
                return None

            token = get_tokens_at_position(node, position)[-1]

            token_and_var = await async_next(
                (
                    (t, v)
                    async for t, v in self.iter_variables_from_token(token, namespace, nodes, position)
                    if position in range_from_token(t)
                ),
                None,
            )

            if (
                token_and_var is None
                and isinstance(node, self.get_expression_statement_types())
                and (token := node.get_token(RobotToken.ARGUMENT)) is not None
                and position in range_from_token(token)
            ):
                token_and_var = await async_next(
                    (
                        (var_token, var)
                        async for var_token, var in self.iter_expression_variables_from_token(
                            token, namespace, nodes, position
                        )
                        if position in range_from_token(var_token)
                    ),
                    None,
                )

            if token_and_var is None:
                return None

            var_token, var = token_and_var
            if var.name == "${CURDIR}":
                return None

            return EvaluatableExpression(range_from_token(var_token), var.name)

        return await run_coroutine_in_thread(run)



from __future__ import annotations

import ast
import asyncio
import time
from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator, List, Optional, cast

from robotcode.core.async_tools import check_canceled, check_canceled_sync, threaded
from robotcode.core.dataclasses import CamelSnakeMixin
from robotcode.core.logging import LoggingDescriptor
from robotcode.core.uri import Uri
from robotcode.jsonrpc2.protocol import rpc_method
from robotcode.language_server.common.lsp_types import DocumentUri, Position, Range, TextDocumentIdentifier
from robotcode.language_server.robotframework.configuration import RobotConfig
from robotcode.language_server.robotframework.utils.async_ast import AsyncVisitor

from .protocol_part import RobotLanguageServerProtocolPart

if TYPE_CHECKING:
    from robotcode.language_server.robotframework.protocol import RobotLanguageServerProtocol


@dataclass
class GetAllTestsParams(CamelSnakeMixin):
    workspace_folder: str
    paths: Optional[List[str]] = None
    suites: Optional[List[str]] = None


@dataclass
class GetTestsParams(CamelSnakeMixin):
    text_document: TextDocumentIdentifier
    base_name: Optional[str]


@dataclass
class GetTestsFromDocumentParams(CamelSnakeMixin):
    text_document: TextDocumentIdentifier


@dataclass
class TestItem(CamelSnakeMixin):
    type: str
    id: str
    label: str
    longname: str
    uri: Optional[DocumentUri] = None
    children: Optional[List[TestItem]] = None
    description: Optional[str] = None
    range: Optional[Range] = None
    tags: Optional[List[str]] = None
    error: Optional[str] = None


class FindTestCasesVisitor(AsyncVisitor):
    async def get(self, source: DocumentUri, model: ast.AST, base_name: Optional[str]) -> List[TestItem]:
        self._results: List[TestItem] = []
        self.source = source
        self.path = Uri(source).to_path().resolve()
        self.base_name = base_name
        await self.visit(model)
        return self._results

    async def visit_Section(self, node: ast.AST) -> None:  # noqa: N802
        from robot.parsing.model.blocks import TestCaseSection

        if isinstance(node, TestCaseSection):
            await self.generic_visit(node)

    async def visit_TestCase(self, node: ast.AST) -> None:  # noqa: N802
        from robot.parsing.model.blocks import TestCase
        from robot.parsing.model.statements import Tags

        test_case = cast(TestCase, node)
        longname = f"{self.base_name}.{test_case.name}" if self.base_name else test_case.name
        self._results.append(
            TestItem(
                type="test",
                id=f"{self.path};{longname};{test_case.lineno}",
                longname=longname,
                label=test_case.name,
                uri=self.source,
                range=Range(
                    start=Position(line=test_case.lineno - 1, character=test_case.col_offset),
                    end=Position(
                        line=(test_case.end_lineno if test_case.end_lineno != -1 else test_case.lineno) - 1,
                        character=test_case.end_col_offset if test_case.end_col_offset != -1 else test_case.col_offset,
                    ),
                ),
                tags=[str(tag) for tag in chain(*[tags.values for tags in test_case.body if isinstance(tags, Tags)])],
            )
        )


class DiscoveringProtocolPart(RobotLanguageServerProtocolPart):
    _logger = LoggingDescriptor()

    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)
        self._patch()

    def _patch(self) -> None:
        from robot.api.parsing import File, get_model
        from robot.running import TestSuite
        from robot.running.builder.builders import RobotParser, TestSuiteBuilder

        from robotcode.language_server.robotframework.utils.version import get_robot_version

        def get_document_text(source: str) -> str:
            check_canceled_sync()

            if self.parent.loop:
                doc = self.parent.documents.get_sync(Uri.from_path(source).normalized())
                if doc is not None:
                    return doc.text()

            return source

        def get_source(self: Any, source: str) -> str:
            return get_document_text(source)

        RobotParser._get_source = get_source

        orig = RobotParser._build

        def my_get_model_v4(source: str, data_only: bool = False, curdir: Optional[str] = None) -> Any:
            check_canceled_sync()
            try:
                return get_model(source, data_only, curdir)
            except (SystemExit, KeyboardInterrupt):
                raise
            except BaseException as e:
                ex = e
                self._logger.critical(lambda: f"Can't parse {source}: {ex}")
                return File(source=source)

        def my_get_model_v6(
            source: str, data_only: bool = False, curdir: Optional[str] = None, lang: Any = None
        ) -> Any:
            check_canceled_sync()
            try:
                return get_model(source, data_only, curdir, lang)
            except (SystemExit, KeyboardInterrupt):
                raise
            except BaseException as e:
                ex = e
                self._logger.critical(lambda: f"Can't parse {source}: {ex}")
                return File(source=source)

        my_get_model = my_get_model_v4 if get_robot_version() < (6, 0) else my_get_model_v6

        def build(
            self: Any,
            suite: TestSuite,
            source: str,
            defaults: Any,
            model: Any = None,
            get_model: Any = my_get_model,
        ) -> TestSuite:
            check_canceled_sync()

            return orig(self, suite, source, defaults, model, get_model)

        RobotParser._build = build

        def _validate_test_counts(self: Any, suite: TestSuite, multisource: bool = False) -> None:
            # we don't need this
            pass

        TestSuiteBuilder._validate_test_counts = _validate_test_counts

    async def get_config(self, workspace_uri: str) -> Optional[RobotConfig]:
        folder = self.parent.workspace.get_workspace_folder(workspace_uri)
        if folder is None:
            return None

        return await self.parent.workspace.get_configuration(RobotConfig, folder.uri)

    @rpc_method(name="robot/discovering/getTestsFromWorkspace", param_type=GetAllTestsParams)
    @threaded()
    async def get_tests_from_workspace(
        self,
        workspace_folder: str,
        paths: Optional[List[str]] = None,
        suites: Optional[List[str]] = None,
        *args: Any,
        **kwargs: Any,
    ) -> List[TestItem]:
        from robot.output.logger import LOGGER
        from robot.running import TestCase, TestSuite
        from robot.running.builder.builders import TestSuiteBuilder

        from robotcode.language_server.robotframework.utils.version import get_robot_version

        async def generate(suite: TestSuite) -> TestItem:
            children: List[TestItem] = []

            await check_canceled()

            test: TestCase
            for test in suite.tests:
                children.append(
                    TestItem(
                        type="test",
                        id=f"{Path(test.source).resolve() if test.source is not None else ''};"
                        f"{test.longname};{test.lineno}",
                        label=test.name,
                        longname=test.longname,
                        uri=str(Uri.from_path(test.source)) if test.source else None,
                        range=Range(
                            start=Position(line=test.lineno - 1, character=0),
                            end=Position(line=test.lineno - 1, character=0),
                        ),
                        tags=list(test.tags),
                    )
                )

            for s in suite.suites:
                children.append(await generate(s))

            return TestItem(
                type="suite",
                id=f"{Path(suite.source).resolve() if suite.source is not None else ''};{suite.longname}",
                label=suite.name,
                longname=suite.longname,
                uri=str(Uri.from_path(suite.source)) if suite.source else None,
                children=children,
                range=Range(
                    start=Position(line=0, character=0),
                    end=Position(line=0, character=0),
                )
                if suite.source
                else None,
            )

        self._logger.info("Wait for workspace documents loaded...")

        await self.parent.robot_workspace.documents_loaded.wait()

        start = time.monotonic()

        self._logger.info("Start discovering tests")

        workspace_path = Uri(workspace_folder).to_path()
        canceled = False
        LOGGER._cache_only = True
        try:
            config = await self.get_config(workspace_folder)
            rpa_mode = config.get_rpa_mode() if config is not None else None
            languages = await self.parent.documents_cache.get_workspace_languages(workspace_folder)
            parsers = config.parsers if config is not None else None

            if paths is None and config is not None:
                paths = config.paths

            if paths and len(paths):

                def normalize_paths(paths: List[str]) -> Iterator[str]:
                    for path in paths:
                        p = Path(path)

                        if not p.is_absolute():
                            p = Path(workspace_path, p)

                        if p.exists():
                            yield str(p)

                def nonexisting_paths(paths: List[str]) -> Iterator[str]:
                    for path in paths:
                        p = Path(path)

                        if not p.is_absolute():
                            p = Path(workspace_path, p)

                        if not p.exists():
                            yield str(p)

                valid_paths = list(normalize_paths(paths))

                if get_robot_version() >= (6, 1):
                    builder = TestSuiteBuilder(
                        included_suites=suites if suites else None,
                        rpa=rpa_mode,
                        lang=languages,
                        parsers=parsers,
                    )
                elif get_robot_version() >= (6, 0):
                    builder = TestSuiteBuilder(
                        included_suites=suites if suites else None,
                        rpa=rpa_mode,
                        lang=languages,
                    )
                else:
                    builder = TestSuiteBuilder(included_suites=suites if suites else None, rpa=rpa_mode)

                suite: Optional[TestSuite] = builder.build(*valid_paths) if valid_paths else None
                suite_item = [await generate(suite)] if suite else []

                return [
                    TestItem(
                        type="workspace",
                        id=str(Path.cwd().resolve()),
                        label=Path.cwd().name,
                        longname=Path.cwd().name,
                        uri=str(Uri.from_path(Path.cwd())),
                        children=[
                            *suite_item,
                            *[
                                TestItem(
                                    type="error",
                                    id=f"{i};ERROR",
                                    longname="error",
                                    label=i,
                                    error=f"Parsing '{i}' failed: File or directory to does not exist.",
                                )
                                for i in nonexisting_paths(paths)
                            ],
                        ],
                    )
                ]

            if get_robot_version() >= (6, 1):
                builder = TestSuiteBuilder(
                    included_suites=suites if suites else None,
                    rpa=rpa_mode,
                    lang=languages,
                    parsers=parsers,
                )
            elif get_robot_version() >= (6, 0):
                builder = TestSuiteBuilder(included_suites=suites if suites else None, rpa=rpa_mode, lang=languages)
            else:
                builder = TestSuiteBuilder(included_suites=suites if suites else None, rpa=rpa_mode)
            return [await generate(builder.build(str(workspace_path)))]
        except (SystemExit, KeyboardInterrupt):
            raise
        except asyncio.CancelledError:
            canceled = True
            self._logger.info("Tests discovery canceled")
            raise
        except BaseException as e:
            ex = e
            self._logger.info(lambda: f"Failed to discover tests: {ex}")
            return [
                TestItem(
                    type="error",
                    id=str(Uri.from_path(Path.cwd().resolve())),
                    longname="error",
                    label=Path.cwd().name,
                    error=str(e),
                )
            ]
        finally:
            if not canceled:
                self._logger.info(lambda: f"Tests discovery took {time.monotonic() - start}s")

    @rpc_method(name="robot/discovering/getTestsFromDocument", param_type=GetTestsParams)
    @threaded()
    async def get_tests_from_document(
        self, text_document: TextDocumentIdentifier, base_name: Optional[str], *args: Any, **kwargs: Any
    ) -> List[TestItem]:
        try:
            return await FindTestCasesVisitor().get(
                text_document.uri,
                await self.parent.documents_cache.get_model(
                    await self.parent.documents.get_or_open_document(
                        Uri(text_document.uri).to_path(), language_id="robotframework"
                    )
                ),
                base_name,
            )
        except (asyncio.CancelledError, SystemExit, KeyboardInterrupt):
            raise
        except BaseException:
            return []
